import logging
from typing import Literal, Optional, Type, Union, Annotated
from pydantic import BaseModel, Field, ConfigDict, field_validator, AliasChoices, Discriminator, FilePath
import datetime
from typing import Dict, Any
from joserfc.jwk import KeySet, RSAKey, OctKey, ECKey
from joserfc import jwt
from joserfc.jwk import GuestProtocol


logger = logging.getLogger(__name__)


class JWTToken(BaseModel):
    """A JWT token

    This is a pydantic model that represents a JWT token.
    It is used to validate the token and to extract information from it.
    The token is decoded using the `decode_token` function.

    """
    model_config = ConfigDict(extra="forbid")

    sub: str
    """A unique identifier for the user (is unique for the issuer)"""
    iss: str
    """The issuer of the token"""
    
    
    exp: datetime.datetime
    """The expiration time of the token"""
    
    
    active_org: str | None = None
    """The active organization of the user, if any"""
    
    client_id: str
    """The client_id of the app that requested the token"""
    preferred_username: str
    """The username of the user"""
    roles: list[str]
    """The roles of the user"""
    scope: str
    """The scope of the token"""
    
    iat: datetime.datetime
    """The issued at time of the token"""

    aud: list[str] | None = None
    """The audience of the token"""
    
    jti: str | None = None
    """The unique identifier for the token"""

    raw: str
    """ The raw original token string """
    
    
    @field_validator("aud", mode="before")
    def aud_to_list(cls: Type["JWTToken"], v: str | list[str] | None) -> list[str] | None:
        """Convert the aud to a list"""
        if not v:
            return None
        if isinstance(v, str):
            return [v]
        
        return v

    @field_validator("sub", mode="before")
    def sub_to_username(cls: Type["JWTToken"], v: str) -> str:
        """Convert the sub to a username compatible string"""
        if isinstance(v, int):
            return str(v)
        return v
    
    @field_validator("iat", mode="before")
    def iat_to_datetime(cls: Type["JWTToken"], v: int) -> datetime.datetime:
        """Convert the iat to a datetime object"""
        if v is None:
            return None
        if isinstance(v, int):
            return datetime.datetime.fromtimestamp(v)
        return v
    
    
    @field_validator("exp", mode="before")
    def exp_to_datetime(cls: Type["JWTToken"], v: int) -> datetime.datetime:
        """Convert the exp to a datetime object"""
        if isinstance(v, int):
            return datetime.datetime.fromtimestamp(v)
        return v

    @property
    def changed_hash(self) -> str:
        """A hash that changes when the user changes"""
        return str(hash(self.sub + self.preferred_username + " ".join(self.roles) + (self.active_org or "")))

    @property
    def scopes(self) -> list[str]:
        """The scopes of the token. Each scope is a string separated by a space"""
        return self.scope.split(" ")
    
    
    def has_scopes(self, scopes: list[str]) -> bool:
        """Check if the user has the given scope"""
        if not scopes:
            return True
        
        return all(scope in self.scopes for scope in scopes)
    
    def has_any_role(self, roles: list[str]) -> bool:
        """Check if the user has any of the given roles"""
        if not roles:
            return True
        
        return any(role in self.roles for role in roles)
        
    
    def has_roles(self, roles: list[str]) -> bool:
        """Check if the user has the given role"""
        if not roles:
            return True
        
        return all(role in self.roles for role in roles)



class StaticToken(JWTToken):
    """A static JWT token"""

    sub: str
    iss: str
    iat: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now()
    )
    exp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now() + datetime.timedelta(days=1)
    )
    client_id: str = "static"
    preferred_username: str = "static_user"
    scope: str = "openid profile email"
    roles: list[str] = Field(default_factory=lambda: ["static"])
    raw: str = Field(default_factory=lambda: "static_token")
    
    

class ImitationRequest(BaseModel):
    """ An imitation request"""
    sub: str
    iss: str
    
   
   
class Issuer(BaseModel):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """
    model_config = ConfigDict(extra="forbid")
    kind: str 
    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    """The issuer of the token""" 
    
    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        raise NotImplementedError("get_jwks not implemented. Must be implemented in subclass")
    
class JWKIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """
    kind: Literal["jwks_dict"] = Field(
        default="jwks_dict",
    )
    

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    """The issuer of the token"""
    
    jwks: Dict[str, Any] = Field(
        validation_alias=AliasChoices("jwks", "JWKS", "JWKS_DICT")
    )
    """The jwks of the issuer"""
    
    
    @field_validator("jwks", mode="before")
    def validate_jwks_dict(cls: Type["JWKIssuer"], v: Dict[str, Any]) -> Dict[str, Any]:
        """Validate the jwks dict"""
        if not isinstance(v, dict):
            raise ValueError("jwks_dict must be a dict")
        if "keys" not in v:
            raise ValueError("jwks_dict must contain a keys field")
        if not isinstance(v["keys"], list):
            raise ValueError("jwks_dict keys must be a list")
        return v
    
    
    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        return self.jwks["keys"]
    
    
    
class RSAKeyIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """
    model_config = ConfigDict(extra="forbid")
    kind: Literal["rsa"] = Field(
        default="rsa",
    )

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    key_id: str = Field(
        default="1",
        validation_alias=AliasChoices("key_id", "kid", "KID")
    )
    """The issuer of the token"""
    public_key: str = Field(
        validation_alias=AliasChoices("public_key", "PUBLIC_KEY")
    )
    
    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        t = RSAKey.import_key(self.public_key)
        return [t.as_dict(kid=self.key_id)]
    
    
    
class RSAKeyFileIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """
    model_config = ConfigDict(extra="forbid")
    kind: Literal["rsa_file"] = Field(
        default="rsa_file",
    )

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    key_id: str = Field(
        default="1",
        validation_alias=AliasChoices("key_id", "kid", "KID")
    )
    """The issuer of the token"""
    public_key_pem_file: FilePath = Field(
        validation_alias=AliasChoices("public_key_pem_file", "PUBLIC_KEY_PEM_FILE")
    )
    
    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        
        with open(self.public_key_pem_file, "rb") as f:
            public_key = f.read()
        
        
        
        t = RSAKey.import_key(public_key)
        return [t.as_dict(kid=self.key_id)]
    
    
    
       
    
    
IssuerUnion = Annotated[Union[JWKIssuer, RSAKeyIssuer, RSAKeyFileIssuer], Discriminator("kind")]


class AuthentikateSettings(BaseModel):
    """The settings for authentikate

    This is a pydantic model that represents the settings for authentikate.
    It is used to configure the library.
    """
    model_config = ConfigDict(extra="forbid")
    
    issuers: list[IssuerUnion] = Field(validation_alias=AliasChoices(
        "issuers", "iss", "issuer", "issuer_url", "ISSUERS", 
    ))
    authorization_headers: list[str] = Field(
        default_factory=lambda: [
            "Authorization",
            "X-Authorization",
            "AUTHORIZATION",
            "authorization",
        ],
        validation_alias=AliasChoices(
            "authorization_headers", "AUTHORIZATION_HEADERS", "AUTHORIZATION_HEADERS"
        )
    )
    static_tokens: dict[str, StaticToken] = Field(default_factory=dict, validation_alias=AliasChoices(
        "static_tokens", "STATIC_TOKENS", "STATIC_TOKENS"
    ))
    """A map of static tokens to their decoded values. Should only be used in tests."""
    
    
    
    
    
    
    def get_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
            
        merged_jwks = {}
        
        for issuer in self.issuers:
            keys = issuer.get_as_jwks()
            
            
            if not isinstance(keys, list):
                raise ValueError("keys must be a list")
            
            for key in keys:
                if key.get("kid") is None:
                    raise ValueError("key must contain a kid field")
                
                if key["kid"] in merged_jwks:
                    raise ValueError(f"Duplicate kid found: {key['kid']}")
                
                merged_jwks[key["kid"]] = key
                
        if not merged_jwks:
            raise ValueError("No keys found in jwks")
        
        
        validated_keys = []
        
        for key in merged_jwks.values():
            validated_keys.append(key)
                
            
            
            
        return validated_keys
    
    
    
    def load_key(self, obj: GuestProtocol) -> KeySet:
        """Resolve the key from the header"""
        kid = obj.headers().get("kid")
        if not kid:
            raise ValueError("Missing kid in header")
        
        
        
        
        key_set = KeySet.import_key_set({"keys": self.get_jwks()})
        return key_set




from authentikate.decode import decode_token
from authentikate.settings import get_settings
from authentikate.base_models import AuthentikateSettings, JWTToken
import re
import logging

logger = logging.getLogger(__name__)  #


def authenticate_token(token: str, settings: AuthentikateSettings) -> JWTToken:
    """
    Authenticate a token and return the auth context
    (containing user, app and scopes)

    """
    decoded: JWTToken

    if token in settings.static_tokens:
        decoded = settings.static_tokens[token]
    else:
        decoded = decode_token(token, settings)

    return decoded


jwt_re = re.compile(r"Bearer\s(?P<token>[^\s]*)")


def extract_plain_from_authorization(authorization: str) -> str:
    """
    Extract a plain token from an Authorization header

    Parameters
    ----------

    authorization : str
        The Authorization header

    Returns
    -------
    str
        The token
    """

    m = jwt_re.match(authorization)
    if m:
        token = m.group("token")
        return token

    raise ValueError("Not a valid token")


def authenticate_header(
    headers: dict[str, str], settings: AuthentikateSettings | None = None
) -> JWTToken:
    """
    Authenticate a request and return the auth context
    (containing user, app and scopes)

    """
    if not settings:
        settings = get_settings()
        
    authorization_header = None

    for i in settings.authorization_headers:
        authorization_header = headers.get(i, None)
        if authorization_header:
            break

    if not authorization_header:
        raise ValueError("No Authorization header")

    token = extract_plain_from_authorization(authorization_header)
    return authenticate_token(token, settings) 
    


def authenticate_header_or_none(
    headers: dict[str, str], settings: AuthentikateSettings | None = None
) -> JWTToken | None:
    """
    Authenticate a request header and return the auth context

    Parameters
    ----------
    headers : dict
        The headers to authenticate

    settings : AuthentikateSettings, optional
        The settings to use, by default None

    Returns
    -------
    Auth | None
        The auth context or None if the token is invalid


    """
    try:
        return authenticate_header(headers, settings)
    except Exception:
        logger.error("Error authenticating header. Skipping!", exc_info=True)
        return None


def authenticate_token_or_none(
    token: str, settings: AuthentikateSettings | None = None
) -> JWTToken | None:
    """
    Authenticate a token and return the auth context

    Tries to authenticate the token, if it fails it will return None


    Parameters
    ----------
    token : str
        The token to authenticate

    settings : AuthentikateSettings, optional
        The settings to use, by default None

    Returns
    -------
    Auth | None
        The auth context or None if the token is invalid


    """

    if not settings:
        settings = get_settings()

    try:
        return authenticate_token(token, settings)
    except Exception:
        logger.error("Error authenticating token. Skipping!", exc_info=True)
        return None







from django.db import models  # Create your models here.
from django.contrib.auth.models import AbstractUser



class Organization(models.Model):
    """An Organization model to represent an organization in the system"""
    slug = models.CharField(max_length=1000, unique=True)

    def __str__(self) -> str:
        """String representation of Organization"""
        return self.slug


class User(AbstractUser):
    """A reflection on the real User"""
    sub = models.CharField(max_length=1000, null=True, blank=True)
    iss = models.CharField(max_length=1000, null=True, blank=True)
    active_organization = models.ForeignKey(
        Organization,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="active_users",
    )
    changed_hash = models.CharField(max_length=1000, null=True, blank=True)

    class Meta:
        """Meta class for User"""

        constraints = [
            models.UniqueConstraint(
                fields=["sub", "iss"],
                condition=models.Q(sub__isnull=False, iss__isnull=False),
                name="unique_sub_iss_if_both_not_null",
            )
        ]
        permissions = [("imitate", "Can imitate me")]


class Client(models.Model):
    """An Oauth2 Client

    An Oauth2 Client is a model to represent an Oauth2 client that is
    registered when a JWT token is authenticated. It retrieves
    the client_id from the token and uses it to create a new
    app or retrieve an existing app. This allows for the grouping
    of users by app.

    """
    iss = models.CharField(max_length=2000, null=True, blank=True)
    client_id = models.CharField(unique=True, max_length=2000)
    name = models.CharField(max_length=2000, null=True, blank=True)

    class Meta:
        """Meta class for Client"""

        unique_together = ("iss", "client_id")

    def __str__(self) -> str:
        """String representation of Client"""
        return f"{self.name}"



from typing import AsyncIterator, Iterator, Union
from strawberry.extensions import SchemaExtension
from kante.context import WsContext, HttpContext
from authentikate.vars import token_var, user_var, client_var, organization_var
from authentikate.base_models import JWTToken
from authentikate.utils import authenticate_token_or_none, authenticate_header_or_none
from authentikate.protocols import UserModel, ClientModel, OrganizationModel
from typing import cast
from authentikate.base_models import AuthentikateSettings

class AuthentikateExtension(SchemaExtension):
    """ This is the extension class for the authentikate extension """
    
    
    def get_settings(self) -> "AuthentikateSettings":
        """ Get the settings for the extension """
        from authentikate.settings import get_settings
        # Call the function to get the settings
        settings = get_settings()
        return settings
    
    
    async def aexpand_user_from_token(self, token: JWTToken) -> "UserModel":
        """ Expand a user from the provided JWT token """
        from authentikate.expand import aexpand_user_from_token
        # Call the async function to expand the user
        user = await aexpand_user_from_token(token)
        return cast(UserModel, user)
    
    async def aexpand_client_from_token(self, token: JWTToken) -> "ClientModel":
        """ Expand a client from the provided JWT token """
        from authentikate.expand import aexpand_client_from_token
        # Call the async function to expand the client
        client = await aexpand_client_from_token(token)
        return cast(ClientModel, client)  
    
    async def aexpand_organization_from_token(self, token: JWTToken) -> "OrganizationModel":
        """ Expand an organization from the provided JWT token """
        from authentikate.expand import aexpand_organization_from_token
        # Call the async function to expand the organization
        organization = await aexpand_organization_from_token(token)
        return cast(OrganizationModel, organization)
        
    
    
    async def on_operation(self) -> Union[AsyncIterator[None], Iterator[None]]:
        """ Set the token in the context variable """
        
        context = self.execution_context.context
        
        reset_user = None
        reset_client = None
        reset_token = None
        
        
        if isinstance(context, WsContext):
            # WebSocket context
            # Do something with the WebSocket context
            
            token = authenticate_token_or_none(
                context.connection_params.get("token", ""),
                self.get_settings(),
            )
            reset_token = token_var.set(token)
            if token:
                user = await self.aexpand_user_from_token(token)
                client = await self.aexpand_client_from_token(token)
                organization = await self.aexpand_organization_from_token(token)
                
                reset_client = client_var.set(client)
                reset_user = user_var.set(user)
                reset_organization = organization_var.set(organization)
                
                context.request.set_user(user)  
                context.request.set_client(client)
                context.request.set_organization(organization)
                context.request.set_extension("token", token)
                
            
            
        
        elif isinstance(context, HttpContext):
            # HTTP context
            # Do something with the HTTP context
            token = authenticate_header_or_none(
                context.headers,
                self.get_settings(),
            )
            reset_token = token_var.set(token)
            if token:
                user = await self.aexpand_user_from_token(token)
                client = await self.aexpand_client_from_token(token)
                organization = await self.aexpand_organization_from_token(token)
                
                reset_client = client_var.set(client)
                reset_user = user_var.set(user)
                reset_organization = organization_var.set(organization)
                
                context.request.set_user(user)  
                context.request.set_client(client)
                context.request.set_organization(organization)
                context.request.set_extension("token", token)
        else:
            raise ValueError("Unknown context type. Cannot determine if it's WebSocket or HTTP.")
           
        
        yield 
        
        
        # Cleanup
        if reset_user:
            user_var.reset(reset_user)
            
        if reset_client:
            client_var.reset(reset_client)
            
        if reset_token:
            token_var.reset(reset_token)
            
        if reset_organization:
            organization_var.reset(reset_organization)
        
        return 
        
       
       

        


from django.apps import AppConfig


class AuthentikateConfig(AppConfig):
    """The Django application configuration for the authentikate app."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "authentikate"



""" Strawberry extension for Authentikate """

from .extension import AuthentikateExtension
from .directives import AuthExtension, Auth, all_directives, AuthSubscribeExtension


__all__ = [
    "AuthentikateExtension"
]



from authentikate import models
import strawberry_django


@strawberry_django.type(models.Organization)
class Organization:
    """ This is the organization type """
    id: str
    slug: str

@strawberry_django.type(models.User)
class User:
    """ This is the user type """
    sub: str
    preferred_username: str
    active_organization: Organization | None = None
    
    
    
@strawberry_django.type(models.Client)
class Client:
    """ This is the client type """
    client_id: str
    name: str


import contextvars
from authentikate.base_models import JWTToken
from authentikate.protocols import UserModel, ClientModel, OrganizationModel



token_var: contextvars.ContextVar[JWTToken | None] = contextvars.ContextVar("token_var", default=None)
user_var: contextvars.ContextVar[UserModel | None] = contextvars.ContextVar("user_var", default=None)
client_var: contextvars.ContextVar[ClientModel | None] = contextvars.ContextVar("client_var", default=None)
organization_var: contextvars.ContextVar[OrganizationModel | None] = contextvars.ContextVar("organization_var", default=None)


def get_token() -> JWTToken | None:
    """
    Get the current token from the context variable

    Returns
    -------
    JWTToken | None
        The current token
    """
    return token_var.get()

        
def get_user() -> UserModel | None:
    """
    Get the current user from the context variable

    Returns
    -------
    User | None
        The current user
    """
    return user_var.get()

def get_client() -> ClientModel | None:
    """
    Get the current client from the context variable

    Returns
    -------
    User | None
        The current user
    """
    return client_var.get()


def get_organization() -> OrganizationModel | None:
    """
    Get the current organization from the context variable

    Returns
    -------
    str | None
        The current organization
    """
    return organization_var.get()
    




